/**
\file model.cpp
\brief Generates a multi-segment OpenSim model.
\author Makoto Yoshida
\date 2023/7/11
\sa https://doi.org/10.1299/jsmec.45.1040
\sa Hyfydy, https://hyfydy.com/
\sa OpenSim 4.4, http://opensim.stanford.edu
\sa CMake, https://cmake.org
\sa Doxygen, https://www.doxygen.nl/
\note Microsoft Visual Studio Community 2019, Release, x64
*/

#include <OpenSim/OpenSim.h>

// right-handed system
// +X is forward
// -X is backward
// +Y is up
// -Y is down
// +Z is right
// -Z is left
// rotation is right-handed screw

/**
\brief Compares floating point values.
\param[in] p1 a floating point value.
\param[in] p2 a floating point value.
\return true if the values are considered equal, otherwise false.
*/
inline bool FuzzyCompare(const double p1, const double p2) {
	return (std::abs(p1 - p2) < SimTK::Eps) ? true : false;
}

/**
\brief Adds a Cylinder.
\param[in,out] model the Model.
\param[in] body the Body to which the Cylinder will be attached.
\param[in] distalEnd the location of the distal end of the Body.*/
void addCylinder(OpenSim::Model& model,
	const OpenSim::Body& body,
	const SimTK::Vec3& distalEnd) {
	auto& pof = *(new OpenSim::PhysicalOffsetFrame());
	pof.setName(body.getName());
	pof.setParentFrame(body);
	pof.setOffsetTransform(SimTK::Transform(distalEnd / 2.));
	pof.attachGeometry(new OpenSim::Cylinder(.02, distalEnd[1] / 2.));
	model.addComponent(&pof);
}

/**
\brief Sets a TransformAxis.
\param[in,out] ta the TransformAxis.
\param[in] name the name of the TransformAxis.
\param[in] axis the axis of the TransformAxis.
*/
void setTransformAxis(OpenSim::TransformAxis& ta,
	const std::string& name, const SimTK::Vec3& axis) {
	ta.setCoordinateNames(OpenSim::Array<std::string>(name, 1));
	ta.setAxis(axis);
	ta.setFunction(new OpenSim::LinearFunction(1., 0.));
}

/**
\brief Adds a CustomJoint.
\param[in,out] model the Model.
\param[in] aName the name of the CustomJoint.
\param[in] parent the parent of the CustomJoint.
\param[in] locationInParent the location of the joint in the parent.
\param[in] child the child of the CustomJoint.
\param[in] r1Name the name of Coordinate r1.
\param[in] r1Axis the axis of Coordinate r1.
\param[in] r1DefaultValue the default value of Coordinate r1.
\param[in] r1RangeMin the minimum value of Coordinate r1.
\param[in] r1RangeMax the maximum value of Coordinate r1.
\param[in] orientationInParent the orientation of the joint in the parent.
*/
void addJoint(OpenSim::Model& model,
	const std::string& aName,
	const OpenSim::PhysicalFrame& parent,
	const SimTK::Vec3& locationInParent,
	const OpenSim::PhysicalFrame& child,
	const std::string& r1Name, const SimTK::Vec3& r1Axis,
	const double r1DefaultValue,
	const double r1RangeMin, const double r1RangeMax,
	const std::string& r2Name, const SimTK::Vec3& r2Axis,
	const double r2DefaultValue,
	const double r2RangeMin, const double r2RangeMax,
	const std::string& r3Name, const SimTK::Vec3& r3Axis,
	const double r3DefaultValue,
	const double r3RangeMin, const double r3RangeMax,
	const std::string& t1Name, const SimTK::Vec3& t1Axis,
	const double t1DefaultValue,
	const double t1RangeMin, const double t1RangeMax,
	const std::string& t2Name, const SimTK::Vec3& t2Axis,
	const double t2DefaultValue,
	const double t2RangeMin, const double t2RangeMax,
	const std::string& t3Name, const SimTK::Vec3& t3Axis,
	const double t3DefaultValue,
	const double t3RangeMin, const double t3RangeMax,
	const SimTK::Vec3& orientationInParent = SimTK::Vec3(0., 0., 0.)) {
	auto& cj = *(new OpenSim::CustomJoint(aName,
		parent, locationInParent, orientationInParent,
		child, SimTK::Vec3(0., 0., 0.), SimTK::Vec3(0., 0., 0.),
		OpenSim::SpatialTransform()));
	cj.append_coordinates(OpenSim::Coordinate(r1Name,
		OpenSim::Coordinate::Rotational,
		r1DefaultValue, r1RangeMin, r1RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(r2Name,
		OpenSim::Coordinate::Rotational,
		r2DefaultValue, r2RangeMin, r2RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(r3Name,
		OpenSim::Coordinate::Rotational,
		r3DefaultValue, r3RangeMin, r3RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(t1Name,
		OpenSim::Coordinate::Translational,
		t1DefaultValue, t1RangeMin, t1RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(t2Name,
		OpenSim::Coordinate::Translational,
		t2DefaultValue, t2RangeMin, t2RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(t3Name,
		OpenSim::Coordinate::Translational,
		t3DefaultValue, t3RangeMin, t3RangeMax));
	setTransformAxis(cj.updSpatialTransform().upd_rotation1(),
		r1Name, r1Axis);
	setTransformAxis(cj.updSpatialTransform().upd_rotation2(),
		r2Name, r2Axis);
	setTransformAxis(cj.updSpatialTransform().upd_rotation3(),
		r3Name, r3Axis);
	setTransformAxis(cj.updSpatialTransform().upd_translation1(),
		t1Name, t1Axis);
	setTransformAxis(cj.updSpatialTransform().upd_translation2(),
		t2Name, t2Axis);
	setTransformAxis(cj.updSpatialTransform().upd_translation3(),
		t3Name, t3Axis);
	model.addJoint(&cj);
}

void addJoint(OpenSim::Model& model,
	const std::string& aName,
	const OpenSim::PhysicalFrame& parent,
	const SimTK::Vec3& locationInParent,
	const OpenSim::PhysicalFrame& child,
	const std::string& r1Name, const SimTK::Vec3& r1Axis,
	const double r1DefaultValue,
	const double r1RangeMin, const double r1RangeMax,
	const std::string& r2Name, const SimTK::Vec3& r2Axis,
	const double r2DefaultValue,
	const double r2RangeMin, const double r2RangeMax,
	const std::string& r3Name, const SimTK::Vec3& r3Axis,
	const double r3DefaultValue,
	const double r3RangeMin, const double r3RangeMax,
	const SimTK::Vec3& orientationInParent = SimTK::Vec3(0., 0., 0.)) {
	auto& cj = *(new OpenSim::CustomJoint(aName,
		parent, locationInParent, orientationInParent,
		child, SimTK::Vec3(0., 0., 0.), SimTK::Vec3(0., 0., 0.),
		OpenSim::SpatialTransform()));
	cj.append_coordinates(OpenSim::Coordinate(r1Name,
		OpenSim::Coordinate::Rotational,
		r1DefaultValue, r1RangeMin, r1RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(r2Name,
		OpenSim::Coordinate::Rotational,
		r2DefaultValue, r2RangeMin, r2RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(r3Name,
		OpenSim::Coordinate::Rotational,
		r3DefaultValue, r3RangeMin, r3RangeMax));
	setTransformAxis(cj.updSpatialTransform().upd_rotation1(),
		r1Name, r1Axis);
	setTransformAxis(cj.updSpatialTransform().upd_rotation2(),
		r2Name, r2Axis);
	setTransformAxis(cj.updSpatialTransform().upd_rotation3(),
		r3Name, r3Axis);
	model.addJoint(&cj);
}

void addJoint(OpenSim::Model& model,
	const std::string& aName,
	const OpenSim::PhysicalFrame& parent,
	const SimTK::Vec3& locationInParent,
	const OpenSim::PhysicalFrame& child,
	const std::string& r1Name, const SimTK::Vec3& r1Axis,
	const double r1DefaultValue,
	const double r1RangeMin, const double r1RangeMax,
	const SimTK::Vec3& orientationInParent = SimTK::Vec3(0., 0., 0.)) {
	auto& cj = *(new OpenSim::CustomJoint(aName,
		parent, locationInParent, orientationInParent,
		child, SimTK::Vec3(0., 0., 0.), SimTK::Vec3(0., 0., 0.),
		OpenSim::SpatialTransform()));
	cj.append_coordinates(OpenSim::Coordinate(r1Name,
		OpenSim::Coordinate::Rotational,
		r1DefaultValue, r1RangeMin, r1RangeMax));
	setTransformAxis(cj.updSpatialTransform().upd_rotation1(),
		r1Name, r1Axis);
	cj.updSpatialTransform().upd_rotation2().setAxis(
		SimTK::Vec3(r1Axis[2], r1Axis[0], r1Axis[1]));
	cj.updSpatialTransform().upd_rotation3().setAxis(
		SimTK::Vec3(r1Axis[1], r1Axis[2], r1Axis[0]));
	model.addJoint(&cj);
}

void addJoint(OpenSim::Model& model,
	const std::string& aName,
	const OpenSim::PhysicalFrame& parent,
	const SimTK::Vec3& locationInParent,
	const OpenSim::PhysicalFrame& child,
	const std::string& r1Name, const SimTK::Vec3& r1Axis,
	const double r1DefaultValue,
	const double r1RangeMin, const double r1RangeMax,
	const std::string& r2Name, const SimTK::Vec3& r2Axis,
	const double r2DefaultValue,
	const double r2RangeMin, const double r2RangeMax,
	const SimTK::Vec3& orientationInParent = SimTK::Vec3(0., 0., 0.)) {
	auto& cj = *(new OpenSim::CustomJoint(aName,
		parent, locationInParent, orientationInParent,
		child, SimTK::Vec3(0., 0., 0.), SimTK::Vec3(0., 0., 0.),
		OpenSim::SpatialTransform()));
	cj.append_coordinates(OpenSim::Coordinate(r1Name,
		OpenSim::Coordinate::Rotational,
		r1DefaultValue, r1RangeMin, r1RangeMax));
	cj.append_coordinates(OpenSim::Coordinate(r2Name,
		OpenSim::Coordinate::Rotational,
		r2DefaultValue, r2RangeMin, r2RangeMax));
	setTransformAxis(cj.updSpatialTransform().upd_rotation1(),
		r1Name, r1Axis);
	setTransformAxis(cj.updSpatialTransform().upd_rotation2(),
		r2Name, r2Axis);
	if (r2Axis != SimTK::Vec3(r1Axis[1], r1Axis[2], r1Axis[0]))
		cj.updSpatialTransform().upd_rotation3().setAxis(
			SimTK::Vec3(r1Axis[1], r1Axis[2], r1Axis[0]));
	else
		cj.updSpatialTransform().upd_rotation3().setAxis(
			SimTK::Vec3(r1Axis[2], r1Axis[0], r1Axis[1]));
	model.addJoint(&cj);
}


/**
\brief Adds a CoordinateLimitForce.
\param[in,out] model the Model.
\param[in] coord the Coordinate.
\param[in] kappa_1 a parameter for the CoordinateLimitForce.
\param[in] kappa_2 a parameter for the CoordinateLimitForce.
\param[in] kappa_3 a parameter for the CoordinateLimitForce.
\param[in] kappa_4 a parameter for the CoordinateLimitForce.
\param[in] kappa_5 a parameter for the CoordinateLimitForce.
\param[in] kappa_6 a parameter for the CoordinateLimitForce.
\param[in] kappa_7 a parameter for the CoordinateLimitForce.
*/
void addCoordinateLimitForce(OpenSim::Model& model,
	const OpenSim::Coordinate& coord,
	const double kappa_1,
	const double kappa_2,
	const double kappa_3,
	const double kappa_4,
	const double kappa_5,
	const double kappa_6,
	const double kappa_7) {
	if (!FuzzyCompare(kappa_1, 0.)) {
		// a rough emulation of n_P, eq. A2

		const double x_1_u = kappa_6;
		const double y_1_u = kappa_4 * std::exp(kappa_5 * (x_1_u - kappa_6));
		const double x_2_u = kappa_6 + SimTK::Pi / 4.;
		const double y_2_u = kappa_4 * std::exp(kappa_5 * (x_2_u - kappa_6));;
		const double K_upper = (y_2_u - y_1_u) / (x_2_u - x_1_u);
		const double q_upper = x_1_u - y_1_u / K_upper;

		const double x_1_l = kappa_3;
		const double y_1_l = kappa_1 * std::exp(kappa_2 * (kappa_3 - x_1_l));
		const double x_2_l = kappa_3 - SimTK::Pi / 4.;
		const double y_2_l = kappa_1 * std::exp(kappa_2 * (kappa_3 - x_2_l));;
		const double K_lower = (y_2_l - y_1_l) / (x_1_l - x_2_l); // negative
		const double q_lower = x_1_l - y_1_l / K_lower;

		model.addComponent(new OpenSim::CoordinateLimitForce(
			coord.getName(),
			SimTK::convertRadiansToDegrees(q_upper + coord.getDefaultValue()),
			std::min(K_upper, 1e3) * SimTK::Pi / 180.,
			SimTK::convertRadiansToDegrees(q_lower + coord.getDefaultValue()),
			std::min(K_lower, 1e3) * SimTK::Pi / 180.,
			kappa_7 * ((std::min(K_upper, 1e3) + std::min(K_lower, 1e3)) * SimTK::Pi / 180. / 2.
				* SimTK::convertDegreesToRadians(5.)) * SimTK::Pi / 180.,
			5.)); // per radians to per degrees
	}
	else {
		if (!FuzzyCompare(kappa_2, 0.))
			model.addComponent(new OpenSim::CoordinateLimitForce(
				coord.getName(),
				SimTK::convertRadiansToDegrees(coord.getDefaultValue()),
				std::min(kappa_2, 1e3) * SimTK::Pi / 180.,
				SimTK::convertRadiansToDegrees(coord.getDefaultValue()),
				std::min(kappa_2, 1e3) * SimTK::Pi / 180.,
				kappa_7 * (std::min(kappa_2, 1e3) * SimTK::Pi / 180.
					* SimTK::convertDegreesToRadians(5.)) * SimTK::Pi / 180.,
				5.)); // per radians to per degrees
	}
}

/**
\brief Adds a CoordinateLimitForce.
\param[in,out] model the Model.
\param[in] name the name of the joint degree of freedom.
\param[in] q_upper the upper limit for the coordinate range of motion.
\param[in] K_upper the stiffness of the passive limit force for the upper limit.
\param[in] q_lower the lower limit for the coordinate range of motion.
\param[in] K_lower the stiffness of the passive limit force for the lower limit.
\param[in] damping a damping factor for the coordinate's speed when limit is exceeded.
*/
void addCoordinateLimitForce(OpenSim::Model& model,
	const std::string& name,
	const double q_upper, 
	const double K_upper,
	const double q_lower, 
	const double K_lower,
	const double damping) {

	model.addComponent(new OpenSim::CoordinateLimitForce(
		name,
		q_upper + SimTK::convertRadiansToDegrees(model.getCoordinateSet().get(name).getDefaultValue()), K_upper,
		q_lower + SimTK::convertRadiansToDegrees(model.getCoordinateSet().get(name).getDefaultValue()), K_lower,
		damping, 5.));
}

/**
\brief Sets a Coordinate at a State.
\param[in,out] model the Model.
\param[in,out] state the State.
\param[in] name the name of the Coordinate.
\param[in] value the value of the Coordinate at the State.
\param[in] speed the speed of the Coordinate at the State.
*/
inline void setCoordinate(OpenSim::Model& model,
	SimTK::State& state,
	const std::string& name,
	const double value,
	const double speed) {
	auto& c = model.updCoordinateSet().get(name);
	c.setValue(state, value);
	c.setSpeedValue(state, speed);
}

/**
\brief Creates a Hyfydy joint component.
\return the joint component to be included in a hfd file.
\sa https://hyfydy.com/documentation/#joint
*/
std::string joint(const std::string& name,
	const std::string& parent,
	const std::string& child,
	const SimTK::Vec3& pos_in_parent,
	const SimTK::Vec3& pos_in_child,
	const SimTK::Vec3& ref_ori,
	const double stiffness,
	const double damping,
	const double limits_xl,
	const double limits_xu,
	const double limits_yl,
	const double limits_yu,
	const double limits_zl,
	const double limits_zu,
	const double limit_stiffness,
	const double limit_damping) {
	std::stringstream ss;

	ss << "\t" << "joint {" << std::endl;
	ss << "\t\t" << "name = " << name << std::endl;
	ss << "\t\t" << "parent = " << parent << std::endl;
	ss << "\t\t" << "child = " << child << std::endl;
	ss << "\t\t" << "pos_in_parent { x = " << pos_in_parent[0]
		<< " y = " << pos_in_parent[1]
		<< " z = " << pos_in_parent[2] << " }" << std::endl;
	ss << "\t\t" << "pos_in_child { x = " << pos_in_child[0]
		<< " y = " << pos_in_child[1]
		<< " z = " << pos_in_child[2] << " }" << std::endl;
	if (ref_ori != SimTK::Vec3(0., 0., 0.))
		ss << "\t\t" << "ref_ori [ "
			<< SimTK::convertRadiansToDegrees(ref_ori[0]) << " "
			<< SimTK::convertRadiansToDegrees(ref_ori[1]) << " "
			<< SimTK::convertRadiansToDegrees(ref_ori[2]) << " ]" << std::endl;
	ss << "\t\t" << "stiffness = " << stiffness << std::endl;
	ss << "\t\t" << "damping = " << damping << std::endl;
	ss << "\t\t" << "limits { x = " << limits_xl << ".." << limits_xu
		<< " y = " << limits_yl << ".." << limits_yu
		<< " z = " << limits_zl << ".." << limits_zu << " }" << std::endl;
	ss << "\t\t" << "limit_stiffness = " << limit_stiffness << std::endl;
	ss << "\t\t" << "limit_damping = " << limit_damping << std::endl;
	ss << "\t" << "}" << std::endl;

	return ss.str();
}

/**
\brief Starts the program.
\param[in] argc the size of the array of the command line arguments.
\param[in] argv the array of the command line arguments.
\return the exit status.
*/
int main(int argc, char *argv[]) {
	OpenSim::Model model;

	model.setAuthors("Makoto Yoshida and Kazunori Hase");
	model.setPublications("https://doi.org/10.1299/jsmec.45.1040");

	model.addBody(new OpenSim::Body("Pelvis", 8.456,
		SimTK::Vec3(0.000, 0.059, 0.000),
		SimTK::Inertia(0.0500, 0.0500, 0.0500)));
	const auto& Pelvis = model.getBodySet().get("Pelvis");
	model.addBody(new OpenSim::Body("Lumbar", 9.468,
		SimTK::Vec3(0.000, 0.090, 0.000),
		SimTK::Inertia(0.0500, 0.0500, 0.0500)));
	const auto& Lumbar = model.getBodySet().get("Lumbar");
	model.addBody(new OpenSim::Body("Thorax", 12.862,
		SimTK::Vec3(0.000, 0.197, 0.000),
		SimTK::Inertia(0.3623, 0.1410, 0.2530)));
	const auto& Thorax = model.getBodySet().get("Thorax");
	model.addBody(new OpenSim::Body("Thigh_r", 5.955,
		SimTK::Vec3(0.000, 0.165, -0.007),
		SimTK::Inertia(0.0866, 0.0219, 0.0908)));
	const auto& Thigh_r = model.getBodySet().get("Thigh_r");
	model.addBody(new OpenSim::Body("LowerLeg_r", 3.632,
		SimTK::Vec3(0.000, 0.182, 0.000),
		SimTK::Inertia(0.0340, 0.0045, 0.0347)));
	const auto& LowerLeg_r = model.getBodySet().get("LowerLeg_r");
	model.addBody(new OpenSim::Body("Thigh_l", 5.955,
		SimTK::Vec3(0.000, 0.165, 0.007),
		SimTK::Inertia(0.0866, 0.0219, 0.0908)));
	const auto& Thigh_l = model.getBodySet().get("Thigh_l");
	model.addBody(new OpenSim::Body("LowerLeg_l", 3.632,
		SimTK::Vec3(0.000, 0.182, 0.000),
		SimTK::Inertia(0.0340, 0.0045, 0.0347)));
	const auto& LowerLeg_l = model.getBodySet().get("LowerLeg_l");
	model.addBody(new OpenSim::Body("UpperArm_r", 1.667,
		SimTK::Vec3(0.000, 0.131, 0.000),
		SimTK::Inertia(0.0105, 0.0018, 0.0109)));
	const auto& UpperArm_r = model.getBodySet().get("UpperArm_r");
	model.addBody(new OpenSim::Body("Forearm_r", 1.310,
		SimTK::Vec3(0.000, 0.108, 0.000),
		SimTK::Inertia(0.0189, 0.0010, 0.0189)));
	const auto& Forearm_r = model.getBodySet().get("Forearm_r");
	model.addBody(new OpenSim::Body("UpperArm_l", 1.667,
		SimTK::Vec3(0.000, 0.131, 0.000),
		SimTK::Inertia(0.0105, 0.0018, 0.0109)));
	const auto& UpperArm_l = model.getBodySet().get("UpperArm_l");
	model.addBody(new OpenSim::Body("Forearm_l", 1.310,
		SimTK::Vec3(0.000, 0.108, 0.000),
		SimTK::Inertia(0.0189, 0.0010, 0.0189)));
	const auto& Forearm_l = model.getBodySet().get("Forearm_l");
	model.addBody(new OpenSim::Body("Head", 4.823,
		SimTK::Vec3(0.000, 0.128, 0.000),
		SimTK::Inertia(0.0322, 0.0322, 0.0322)));
	const auto& Head = model.getBodySet().get("Head");


	addCylinder(model, Pelvis, SimTK::Vec3(0.000, 0.220, 0.000));
	addCylinder(model, Lumbar, SimTK::Vec3(0.000, 0.180, 0.000));
	addCylinder(model, Thorax, SimTK::Vec3(0.000, 0.240, 0.000));
	addCylinder(model, Thigh_r, SimTK::Vec3(0.000, 0.380, -0.030));
	addCylinder(model, LowerLeg_r, SimTK::Vec3(0.000, 0.420, 0.000));
	addCylinder(model, Thigh_l, SimTK::Vec3(0.000, 0.380, 0.030));
	addCylinder(model, LowerLeg_l, SimTK::Vec3(0.000, 0.420, 0.000));
	addCylinder(model, UpperArm_r, SimTK::Vec3(0.000, 0.300, 0.000));
	addCylinder(model, Forearm_r, SimTK::Vec3(0.000, 0.250, 0.000));
	addCylinder(model, UpperArm_l, SimTK::Vec3(0.000, 0.300, 0.000));
	addCylinder(model, Forearm_l, SimTK::Vec3(0.000, 0.250, 0.000));
	addCylinder(model, Head, SimTK::Vec3(0.000, 0.200, 0.000));

	addJoint(model, "global",
		model.getGround(), SimTK::Vec3(0., 0., 0.),
		Pelvis,
		"yaw", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"roll", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"pitch", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(-10.00), -SimTK::Pi, SimTK::Pi,
		"longitudinal", SimTK::Vec3(1, 0, 0),
		0., -10., 10.,
		"vertical", SimTK::Vec3(0, 1, 0),
		0., -10., 10.,
		"lateral", SimTK::Vec3(0, 0, 1),
		0., -10., 10.);
	addJoint(model, "SpineL",
		Pelvis,
		SimTK::Vec3(0.000, 0.220, 0.000),
		Lumbar,
		"SpineL_LL", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"SpineL_RR", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"SpineL_FE", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(18.00), -SimTK::Pi, SimTK::Pi);
	// SpineL_LL, left lateral bending < right lateral bending
	// SpineL_RR, right rotation < left rotation
	// SpineL_FE, flexion < extension
	addJoint(model, "SpineU",
		Lumbar,
		SimTK::Vec3(0.000, 0.180, 0.000),
		Thorax,
		"SpineU_FE", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(-9.00), -SimTK::Pi, SimTK::Pi);
	// SpineU_FE, flexion < extension
	addJoint(model, "Hip_r",
		Pelvis,
		SimTK::Vec3(0.000, 0.000, 0.100),
		Thigh_r,
		"Hip_EF_r", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(180. - 181.54 + 10.00), -SimTK::Pi, SimTK::Pi,
		"Hip_RR_r", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"Hip_DB_r", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		SimTK::Vec3(0., 0., SimTK::Pi)); // Hyfydy
	// Hip_EF_r, extension < flexion, pi
	// Hip_RR_r, internal rotation < external rotation
	// Hip_DB_r, adduction < abduction
	addJoint(model, "Knee_r",
		Thigh_r,
		SimTK::Vec3(0.000, 0.380, -0.013),
		LowerLeg_r,
		"Knee_FE_r", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi);
	// Knee_FE_r, flexion < extension
	addJoint(model, "Hip_l",
		Pelvis,
		SimTK::Vec3(0.000, 0.000, -0.100),
		Thigh_l,
		"Hip_EF_l", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(180. - 181.54 + 10.00), -SimTK::Pi, SimTK::Pi,
		"Hip_RR_l", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"Hip_BD_l", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		SimTK::Vec3(0., 0., SimTK::Pi)); // Hyfydy
	// Hip_EF_l, extension < flexion, pi
	// Hip_RR_l, external rotation < internal rotation
	// Hip_BD_l, abduction < adduction
	addJoint(model, "Knee_l",
		Thigh_l,
		SimTK::Vec3(0.000, 0.380, 0.013),
		LowerLeg_l,
		"Knee_FE_l", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi);
	// Knee_FE_l, flexion < extension
	addJoint(model, "Shoulder_r",
		Thorax,
		SimTK::Vec3(0.000, 0.120, 0.198),
		UpperArm_r,
		"Shoulder_BD_r", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"Shoulder_EF_r", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(180. - 180.00), -SimTK::Pi, SimTK::Pi,
		"Shoulder_RR_r", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		SimTK::Vec3(0., 0., SimTK::Pi)); // Hyfydy
	// Shoulder_BD_r, abduction < adduction
	// Shoulder_EF_r, extension < flexion, pi
	// Shoulder_RR_r, external rotation < internal rotation
	addJoint(model, "Elbow_r",
		UpperArm_r,
		SimTK::Vec3(0.000, 0.300, 0.000),
		Forearm_r,
		"Elbow_EF_r", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi);
	// Elbow_EF_r, extension < flexion
	addJoint(model, "Shoulder_l",
		Thorax,
		SimTK::Vec3(0.000, 0.120, -0.198),
		UpperArm_l,
		"Shoulder_DB_l", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		"Shoulder_EF_l", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(180. - 180.00), -SimTK::Pi, SimTK::Pi,
		"Shoulder_RR_l", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi,
		SimTK::Vec3(0., 0., SimTK::Pi)); // Hyfydy
	// Shoulder_DB_l, adduction < abduction
	// Shoulder_EF_l, extension < flexion, pi
	// Shoulder_RR_l, internal rotation < external rotation
	addJoint(model, "Elbow_l",
		UpperArm_l,
		SimTK::Vec3(0.000, 0.300, 0.000),
		Forearm_l,
		"Elbow_EF_l", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(0.00), -SimTK::Pi, SimTK::Pi);
	// Elbow_EF_l, extension < flexion
	addJoint(model, "Neck",
		Thorax,
		SimTK::Vec3(0.000, 0.240, 0.000),
		Head,
		"Neck_LL", SimTK::Vec3(1, 0, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::convertDegreesToRadians(35.00), SimTK::convertDegreesToRadians(35.00),
		"Neck_RR", SimTK::Vec3(0, 1, 0),
		SimTK::convertDegreesToRadians(0.00), -SimTK::convertDegreesToRadians(70.00), SimTK::convertDegreesToRadians(70.00),
		"Neck_FE", SimTK::Vec3(0, 0, 1),
		SimTK::convertDegreesToRadians(1.00), -SimTK::convertDegreesToRadians(55.00), SimTK::Pi);
	// Neck_LL, left lateral bending < right lateral bending
	// Neck_RR, right rotation < left rotation
	// Neck_FE, flexion < extension

	addCoordinateLimitForce(model,
		model.getCoordinateSet().get("yaw"),
		0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000);
	addCoordinateLimitForce(model,
		model.getCoordinateSet().get("roll"),
		0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000);
	addCoordinateLimitForce(model,
		model.getCoordinateSet().get("pitch"),
		0.000, 0.000, 0.000, 0.000, 0.000, 0.000, 0.000);
	addCoordinateLimitForce(model,
		model.getCoordinateSet().get("SpineL_LL"),
		1.000, 15.000, 0.000, 1.000, 15.000, 0.000, 3.000);
	addCoordinateLimitForce(model,
		model.getCoordinateSet().get("SpineL_RR"),
		1.000, 15.000, 0.000, 1.000, 15.000, 0.000, 3.000);

	addCoordinateLimitForce(model, "SpineL_FE", 45, 1000, -45, 1000, 0.05);
	addCoordinateLimitForce(model, "SpineU_FE", 40, 1000, -30, 1000, 0.05);
	addCoordinateLimitForce(model, "Hip_EF_r", 133.8, 0.34, -18.1, 0.55, 0.05);
	addCoordinateLimitForce(model, "Hip_RR_r", 40, 114, -40, 114, 0.05);
	addCoordinateLimitForce(model, "Hip_DB_r", 50, 0.64, -30, 0.64, 0.05);
	addCoordinateLimitForce(model, "Knee_FE_r", 1.6, 6000000, -141.9, 7, 0.05);
	addCoordinateLimitForce(model, "Hip_EF_l", 133.8, 0.34, -18.1, 0.55, 0.05);
	addCoordinateLimitForce(model, "Hip_RR_l", 40, 114, -40, 114, 0.05);
	addCoordinateLimitForce(model, "Hip_BD_l", 30, 0.64, -50, 0.64, 0.05);
	addCoordinateLimitForce(model, "Knee_FE_l", 1.6, 6000000, -141.9, 7, 0.05);
	addCoordinateLimitForce(model, "Shoulder_BD_r", 30, 0.08, -150, 0.08, 0.05);
	addCoordinateLimitForce(model, "Shoulder_EF_r", 172, 0.09, -172, 0.09, 0.05);
	addCoordinateLimitForce(model, "Shoulder_RR_r", 50, 114, -90, 114, 0.05);
	addCoordinateLimitForce(model, "Elbow_EF_r", 150, 3.5, -4.7, 1235, 0.05);
	addCoordinateLimitForce(model, "Shoulder_DB_l", 150, 0.08, -30, 0.08, 0.05);
	addCoordinateLimitForce(model, "Shoulder_EF_l", 172, 0.09, -172, 0.09, 0.05);
	addCoordinateLimitForce(model, "Shoulder_RR_l", 50, 114, -90, 114, 0.05);
	addCoordinateLimitForce(model, "Elbow_EF_l", 150, 3.5, -4.7, 1235, 0.05);
	addCoordinateLimitForce(model, "Neck_LL", 30, 1000, -30, 1000, 0.05);
	addCoordinateLimitForce(model, "Neck_RR", 90, 1000, -90, 1000, 0.05);
	addCoordinateLimitForce(model, "Neck_FE", 80, 10, -70, 10, 0.05);

	// cuboid
	const double bpMass = 4.; // kg
	const double bpX = .1; // m
	const double bpY = .4; // m
	const double bpZ = .3; // m


	// To make LuaDof:is_actuated() true.
	for (int i = 0; i < model.getNumCoordinates(); ++i) {
		const auto& c = model.getCoordinateSet().get(i);
		if (c.getName() == "yaw"
			|| c.getName() == "roll"
			|| c.getName() == "pitch"
			|| c.getName() == "longitudinal"
			|| c.getName() == "vertical"
			|| c.getName() == "lateral")
			continue;
		auto& ca = *(new OpenSim::CoordinateActuator(c.getName()));
		ca.setName(c.getName() + "_ca");
		ca.setOptimalForce(1.);
		ca.setMinControl(-1e3);
		ca.setMaxControl(1e3);
		model.addForce(&ca);
	}

	SimTK::State state = model.initSystem();
	model.realizePosition(state);
	for (int i = 0; i < model.getMuscles().getSize(); ++i) {
		auto& m = model.updMuscles().get(i);
		m.setOptimalFiberLength(m.getLength(state) * .9);
		m.setTendonSlackLength(m.getLength(state) * .1);
	}
	model.print("model.osim");

	std::vector<double> q_bar(model.getNumCoordinates());
	for (int i = 0; i < model.getNumCoordinates(); ++i)
		q_bar.at(i) = model.getCoordinateSet().get(i).getDefaultValue();

	std::vector<SimTK::Vec3> theta_bar(model.getNumBodies());

	model.realizePosition(state);
	for (int i = 0; i < model.getNumBodies(); ++i) {
		const auto& b = model.getBodySet().get(i);
		theta_bar.at(i) = b.getRotationInGround(state).convertRotationToBodyFixedXYZ();
		if ((b.getName() == "Thigh_r"
			|| b.getName() == "LowerLeg_r"
			|| b.getName() == "Foot_r"
			|| b.getName() == "Thigh_l"
			|| b.getName() == "LowerLeg_l"
			|| b.getName() == "Foot_l"
			|| b.getName() == "UpperArm_r"
			|| b.getName() == "Forearm_r"
			|| b.getName() == "UpperArm_l"
			|| b.getName() == "Forearm_l")) {
			if (theta_bar.at(i)[2] < 0.)
				theta_bar.at(i)[2] += 2. * SimTK::Pi;
			theta_bar.at(i)[2] -= SimTK::Pi;
		}
	}

	// https://github.com/tgeijten/scone-core
	// https://github.com/tgeijten/scone-core/blob/main/src/sconelib/scone/core/ModelConverter.cpp
	{
		// A partial Hyfydy model to be manually merged to an automatically converted Hyfydy model.
		// Remove every joint components included in body components except those without CoordinateLimitForce.
		// Transplant every mesh components included in body components.

		std::ofstream ofs("model_hfd.hfd");
		ofs << "model {" << std::endl;

		for (auto& c : model.getComponentList<OpenSim::Cylinder>()) {
			ofs << "\t" << "body {" << std::endl;
			ofs << "\t\t" << "name = " << c.getFrame().findBaseFrame().getName() << std::endl;
			ofs << "\t\t" << "mesh {" << std::endl;
			ofs << "\t\t\t" << "shape {" << std::endl;
			ofs << "\t\t\t\t" << "type = cylinder" << std::endl;
			ofs << "\t\t\t\t" << "height = " << c.get_half_height() * 2. << std::endl;
			ofs << "\t\t\t\t" << "radius = " << c.get_radius() << std::endl;
			ofs << "\t\t\t" << "}" << std::endl;
			ofs << "\t\t" << "}" << std::endl;
			ofs << "\t" << "}" << std::endl;
		}

		for (auto& b : model.getComponentList<OpenSim::Brick>()) {
			ofs << "\t" << "body {" << std::endl;
			ofs << "\t\t" << "name = " << b.getFrame().findBaseFrame().getName() << std::endl;
			ofs << "\t\t" << "mesh {" << std::endl;
			ofs << "\t\t\t" << "shape {" << std::endl;
			ofs << "\t\t\t\t" << "type = box" << std::endl;
			ofs << "\t\t\t\t" << "dim = [ "
				<< b.getProperty_half_lengths().getValue()[0] * 2. << " "
				<< b.getProperty_half_lengths().getValue()[1] * 2. << " "
				<< b.getProperty_half_lengths().getValue()[2] * 2. << " ]" << std::endl;
			ofs << "\t\t\t" << "}" << std::endl;
			ofs << "\t\t" << "}" << std::endl;
			ofs << "\t" << "}" << std::endl;
		} // Set ref_ori in the joint appropriately.

		std::map<std::string, std::string> m;
		for (const auto& clf : model.getComponentList<OpenSim::CoordinateLimitForce>()) {
			const auto& c = model.getCoordinateSet().get(clf.get_coordinate());
			const auto& cj = *OpenSim::CustomJoint::safeDownCast(const_cast<OpenSim::Joint*>(&c.getJoint()));
			const auto& pofp = *OpenSim::PhysicalOffsetFrame::safeDownCast(const_cast<OpenSim::PhysicalFrame*>(&cj.getParentFrame()));
			const auto& pofc = *OpenSim::PhysicalOffsetFrame::safeDownCast(const_cast<OpenSim::PhysicalFrame*>(&cj.getChildFrame()));
			const auto& bp = *OpenSim::Body::safeDownCast(const_cast<OpenSim::PhysicalFrame*>(&pofp.getParentFrame()));
			const auto& bc = *OpenSim::Body::safeDownCast(const_cast<OpenSim::PhysicalFrame*>(&pofc.getParentFrame()));
			const auto& parent = bp.getName();
			const auto& child = bc.getName();
			const auto& pos_in_parent = pofp.get_translation() - bp.getMassCenter();
			const auto& pos_in_child = pofc.get_translation() - bc.getMassCenter();
			const auto& ref_ori = pofp.get_orientation();
			const auto& st = cj.getSpatialTransform();
			auto axis = SimTK::Vec3(0., 0., 0.);
			for (int j = 0; j < st.getCoordinateNames().size(); ++j)
				if (st.getCoordinateNames()[j] == c.getName())
					axis = st.getAxes().at(j);

			if (axis == SimTK::Vec3(1., 0., 0.)) {
				ofs << joint(c.getName() + "_Lower", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					clf.getLowerLimit(), 360., -360., 360., -360., 360.,
					clf.getLowerStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
				ofs << joint(c.getName() + "_Upper", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					-360., clf.getUpperLimit(), -360., 360., -360., 360.,
					clf.getUpperStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
			}

			if (axis == SimTK::Vec3(0., 1., 0.)) {
				ofs << joint(c.getName() + "_Lower", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					-360., 360., clf.getLowerLimit(), 360., -360., 360.,
					clf.getLowerStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
				ofs << joint(c.getName() + "_Upper", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					-360., 360., -360., clf.getUpperLimit(), -360., 360.,
					clf.getUpperStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
			}

			if (axis == SimTK::Vec3(0., 0., 1.)) {
				ofs << joint(c.getName() + "_Lower", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					-360., 360., -360., 360., clf.getLowerLimit(), 360.,
					clf.getLowerStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
				ofs << joint(c.getName() + "_Upper", parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					0., 0.,
					-360., 360., -360., 360., -360., clf.getUpperLimit(),
					clf.getUpperStiffness() * 180. / SimTK::Pi,
					clf.getDamping() / ((clf.getLowerStiffness() + clf.getUpperStiffness()) / 2. * SimTK::convertDegreesToRadians(5.)) * 180. / SimTK::Pi);
			}

			m.insert(std::make_pair(cj.getName(),
				joint(cj.getName(), parent, child,
					pos_in_parent, pos_in_child, ref_ori,
					1e6, 500.,
					-360., 360., -360., 360., -360., 360.,
					1e6, 500.)));
			// Manually fix the dof that should be fixed by setting the range to 0..0 instead of -360..360.
			// Manually remove the limit_stiffnes property and the limit_damping property.
		}

		// has to be here.
		for (int i = 0; i < model.getNumJoints(); ++i) {
			const auto& n = model.getJointSet().get(i).getName();
			if (m.count(n) != 0) ofs << m.at(n);
		}

		m.clear();
		for (int i = 0; i < model.getForceSet().getSize(); ++i) {
			const auto& f = model.getForceSet().get(i);
			if (f.getConcreteClassName() == "CoordinateActuator") {
				const auto& ca = *OpenSim::CoordinateActuator::safeDownCast(
					const_cast<OpenSim::Force*>(&f));
				const auto& n = ca.getCoordinate()->getJoint().getName();
				std::stringstream ss;
				ss << "\t" << "joint_motor {" << std::endl;
				ss << "\t\t" << "joint = " << n << std::endl;
				ss << "\t\t" << "stiffness = 0" << std::endl;
				ss << "\t\t" << "damping = 0" << std::endl;
				ss << "\t\t" << "max_torque = 1000" << std::endl;
				ss << "\t" << "}" << std::endl;
				m.insert(std::make_pair(n, ss.str()));
			}
		}
		for (int i = 0; i < model.getNumJoints(); ++i) {
			const auto& n = model.getJointSet().get(i).getName();
			if (m.count(n) != 0) ofs << m.at(n);
		}

		ofs << "}" << std::endl;
		ofs.close();
	}

	{
		// Trailing commas are valid.
		std::ofstream ofs("model.lua");
		ofs << "return {" << std::endl;
		ofs << "\t" << "q_bar = {" << std::endl;
		for (int i = 0; i < model.getNumCoordinates(); ++i)
			ofs << "\t\t" << model.getCoordinateSet().get(i).getName() << " = "
				<< q_bar.at(i) << "," << std::endl;
		ofs << "\t}," << std::endl;
		ofs << "\t" << "CoordinateLimitForce = {" << std::endl;
		for (const auto& clf : model.getComponentList<OpenSim::CoordinateLimitForce>())
			ofs << "\t\t" << clf.get_coordinate() << " = "
				<< SimTK::convertDegreesToRadians(
					clf.getUpperLimit() - clf.getLowerLimit()) << "," << std::endl;
		ofs << "\t}," << std::endl;
		ofs << "\t" << "theta_bar = {" << std::endl;
		for (int i = 0; i < model.getNumBodies(); ++i)
			ofs << "\t\t" << model.getBodySet().get(i).getName() << " = { "
				<< "x = " << theta_bar.at(i)[0] << ", "
				<< "y = " << theta_bar.at(i)[1] << ", "
				<< "z = " << theta_bar.at(i)[2]
				<< " }," << std::endl;
		ofs << "\t}," << std::endl;
		ofs << "\t" << "ContactGeometrySet = {" << std::endl;
		for (int i = 0; i < model.getNumContactGeometries(); ++i) {
			const auto& cg = model.getContactGeometrySet().get(i);
			ofs << "\t\t" << cg.getName() << " = { "
				<< "body = \"" << cg.getBody().getName() << "\", "
				<< "x = " << cg.getLocation()[0] << ", "
				<< "y = " << cg.getLocation()[1] << ", "
				<< "z = " << cg.getLocation()[2]
				<< " }," << std::endl;
		}
		ofs << "\t}," << std::endl;
		ofs << "}" << std::endl;
		ofs.close();
	}

	state.setTime(0.);

	setCoordinate(model, state, "yaw",
		SimTK::convertDegreesToRadians(-1.031),
		SimTK::convertDegreesToRadians(-0.726));
	setCoordinate(model, state, "roll",
		SimTK::convertDegreesToRadians(-0.608),
		SimTK::convertDegreesToRadians(-2.109));
	setCoordinate(model, state, "pitch",
		SimTK::convertDegreesToRadians(-2.071),
		SimTK::convertDegreesToRadians(-2.094));
	setCoordinate(model, state, "SpineL_LL",
		SimTK::convertDegreesToRadians(1.184),
		SimTK::convertDegreesToRadians(-1.305));
	setCoordinate(model, state, "SpineL_RR",
		SimTK::convertDegreesToRadians(-1.299),
		SimTK::convertDegreesToRadians(23.655));
	setCoordinate(model, state, "SpineL_FE",
		SimTK::convertDegreesToRadians(-19.458),
		SimTK::convertDegreesToRadians(-14.781));
	setCoordinate(model, state, "SpineM_FE",
		SimTK::convertDegreesToRadians(31.479),
		SimTK::convertDegreesToRadians(-5.455));
	setCoordinate(model, state, "SpineU_FE",
		SimTK::convertDegreesToRadians(-15.157),
		SimTK::convertDegreesToRadians(0.312));
	setCoordinate(model, state, "Hip_EF_r",
		SimTK::convertDegreesToRadians(180. - 175.911),
		SimTK::convertDegreesToRadians(-56.234)); // Hyfydy
	setCoordinate(model, state, "Hip_RR_r",
		SimTK::convertDegreesToRadians(-5.037),
		SimTK::convertDegreesToRadians(-4.169));
	setCoordinate(model, state, "Hip_DB_r",
		SimTK::convertDegreesToRadians(-1.651),
		SimTK::convertDegreesToRadians(-1.882));
	setCoordinate(model, state, "Knee_FE_r",
		SimTK::convertDegreesToRadians(-3.037),
		SimTK::convertDegreesToRadians(-14.671));
	setCoordinate(model, state, "Ankle_IE_r",
		SimTK::convertDegreesToRadians(-0.437),
		SimTK::convertDegreesToRadians(-7.781));
	setCoordinate(model, state, "Ankle_PD_r",
		SimTK::convertDegreesToRadians(-3.742),
		SimTK::convertDegreesToRadians(50.320));
	setCoordinate(model, state, "Hip_EF_l",
		SimTK::convertDegreesToRadians(180. - 153.886),
		SimTK::convertDegreesToRadians(189.784)); // Hyfydy
	setCoordinate(model, state, "Hip_RR_l",
		SimTK::convertDegreesToRadians(0.894),
		SimTK::convertDegreesToRadians(-30.085));
	setCoordinate(model, state, "Hip_BD_l",
		SimTK::convertDegreesToRadians(2.775),
		SimTK::convertDegreesToRadians(-18.430));
	setCoordinate(model, state, "Knee_FE_l",
		SimTK::convertDegreesToRadians(-73.388),
		SimTK::convertDegreesToRadians(63.241));
	setCoordinate(model, state, "Ankle_EI_l",
		SimTK::convertDegreesToRadians(0.242),
		SimTK::convertDegreesToRadians(23.869));
	setCoordinate(model, state, "Ankle_PD_l",
		SimTK::convertDegreesToRadians(-5.373),
		SimTK::convertDegreesToRadians(26.729));
	setCoordinate(model, state, "Shoulder_BD_r",
		SimTK::convertDegreesToRadians(1.925),
		SimTK::convertDegreesToRadians(3.500));
	setCoordinate(model, state, "Shoulder_EF_r",
		SimTK::convertDegreesToRadians(180. - 176.120),
		SimTK::convertDegreesToRadians(50.580)); // Hyfydy
	setCoordinate(model, state, "Elbow_EF_r",
		SimTK::convertDegreesToRadians(8.770),
		SimTK::convertDegreesToRadians(-32.381));
	setCoordinate(model, state, "Shoulder_DB_l",
		SimTK::convertDegreesToRadians(4.362),
		SimTK::convertDegreesToRadians(1.505));
	setCoordinate(model, state, "Shoulder_EF_l",
		SimTK::convertDegreesToRadians(180. - 177.158),
		SimTK::convertDegreesToRadians(16.370)); // Hyfydy
	setCoordinate(model, state, "Elbow_EF_l",
		SimTK::convertDegreesToRadians(11.404),
		SimTK::convertDegreesToRadians(24.644));
	setCoordinate(model, state, "longitudinal", 0.000, 0.989);
	setCoordinate(model, state, "lateral", 0.000, 0.006);
	setCoordinate(model, state, "vertical", 0.861, 0.059);

	OpenSim::Storage r;
	OpenSim::Array<std::string> cl;
	cl.append("time");
	for (int i = 0; i < model.getStateVariableNames().size(); ++i)
		cl.append(model.getStateVariableNames().get(i));

	r.setColumnLabels(cl);
	OpenSim::Array<double> v;
	for (int i = 0; i < model.getStateVariableNames().size(); ++i)
		v.append(model.getStateVariableValue(state,
			model.getStateVariableNames().get(i)));
	r.append(state.getTime(), v);
	r.print("initial.sto");

	{
		std::ofstream ofs("initial.zml");
		ofs << "values {" << std::endl;
		for (int i = 0; i < cl.size(); ++i) {
			const std::string& s = cl[i];
			std::smatch m;
			if (std::regex_search(s, m, std::regex(R"(/jointset/(\w+)/(\w+)/value)")))
				ofs << "\t" << m[2] << " = "
					<< model.getStateVariableValue(state, s) << std::endl;
		}
		ofs << "}" << std::endl;
		ofs << "velocities {" << std::endl;
		for (int i = 0; i < cl.size(); ++i) {
			const std::string& s = cl[i];
			std::smatch m;
			if (std::regex_search(s, m, std::regex(R"(/jointset/(\w+)/(\w+)/speed)")))
				ofs << "\t" << m[2] << " = "
					<< model.getStateVariableValue(state, s) << std::endl;
		}
		ofs << "}" << std::endl;
		ofs.close();
	}

	return EXIT_SUCCESS;
}
