/**
\file kinect.cpp
\brief Connects Microsoft Azure Kinect DK and OpenSim.
\author Adelina Baerligea
\author Makoto Yoshide
\date 2023/9/26
\sa Microsoft Azure Kinect DK, https://learn.microsoft.com/en-us/azure/kinect-dk/
\sa Azure Kinect Sensor SDK
\sa Azure Kinect Body Tracking SDK
\sa Matroska, https://matroska.org/
\sa OpenSim 4.4, http://opensim.stanford.edu
\sa CMake, https://cmake.org
\sa Doxygen, https://www.doxygen.nl/
*/

#include <OpenSim/OpenSim.h>
#include <k4a/k4a.hpp>
#include <k4abt.hpp>
#include <k4arecord/playback.hpp>

/*
USB 3.0
nVidia GeForce GTX 1050

+X is right
+Y is down
+Z is forward
*/

// additional functions for coordinate system transformations
// from https://github.com/andyj1/kinect
#include "transform.h"

/**
\brief Returns the discrete confidence_level of a joint.
\param[in] n the name of the joint.
\param[in] j the joint.
*/
double confidence_level(const k4abt_joint_t& j) {
	switch (j.confidence_level) {
	case K4ABT_JOINT_CONFIDENCE_NONE:
		return 0;
	case K4ABT_JOINT_CONFIDENCE_LOW:
		return 1;
	case K4ABT_JOINT_CONFIDENCE_MEDIUM:
		return 2;
	}
}

/**
\brief Converts two mkv files to trc files, transforming the coordinates of the second [2] to match the first [1].
\param[in] fnMKV the name of the Matroska file - [1,2].
\param[in, out] tstv the data table (vec3) for marker movement - [1,2].
\param[in, out] tst the data table for marker confidence level - [1,2].
*/
void tracking(const std::string& fnMKV1, const std::string& fnMKV2,
	OpenSim::TimeSeriesTableVec3& tstv1, OpenSim::TimeSeriesTableVec3& tstv2,
	OpenSim::TimeSeriesTable& tst1, OpenSim::TimeSeriesTable& tst2) {

	auto playback1 = k4a::playback::open(fnMKV1.c_str());
	auto playback2 = k4a::playback::open(fnMKV2.c_str());
	std::cout << "Calibration start." << std::endl;
	auto tracker1 = k4abt::tracker::create(playback1.get_calibration());
	auto tracker2 = k4abt::tracker::create(playback2.get_calibration());
	std::cout << "Calibration end." << std::endl;

	k4a::capture capture1, capture2;
	std::cout << "Capture start." << std::endl;

	while (playback1.get_next_capture(&capture1)) {
		playback2.get_next_capture(&capture2);

		if (!tracker1.enqueue_capture(capture1)) break;
		capture1.reset();
		if (!tracker2.enqueue_capture(capture2)) break;
		capture2.reset();

		k4abt::frame frame1, frame2;
		if (!tracker1.pop_result(&frame1)) break;
		if (!tracker2.pop_result(&frame2)) break;
		std::cout << ((double)(frame1.get_device_timestamp().count())) / 1e6 << " s" << std::endl;

		uint32_t main_num_bodies = 0;
		uint32_t secondary_num_bodies = 0;

		if ((frame1 != nullptr) && (frame2 != nullptr))
		{
			main_num_bodies = frame1.get_num_bodies();
			secondary_num_bodies = frame2.get_num_bodies();
		}

		if (main_num_bodies > 0 && secondary_num_bodies > 0) 
		{
			if (std::max(main_num_bodies, secondary_num_bodies) > 1)
			{
				std::cout << "Warning! More than one body detected by the camera!" <<
					"This can lead to errors in the tracking process!" << std::endl;
			}

			k4abt_body_t main_body = frame1.get_body(0);
			k4abt_body_t secondary_body = frame2.get_body(0);

			// capture confidence levels
			std::vector<double> c1(K4ABT_JOINT_COUNT), c2(K4ABT_JOINT_COUNT);
			for (int i = 0; i < (int)K4ABT_JOINT_COUNT; ++i)
			{
				c1[i] = confidence_level(main_body.skeleton.joints[i]);
				c2[i] = confidence_level(secondary_body.skeleton.joints[i]);
			}

			// transform secondary to main body coordinate system
			try 
			{
				transform_body(main_body, secondary_body);
			}
			catch (std::domain_error& e)
			{
				std::cerr << e.what() << std::endl;
				for (int i = 0; i < (int)K4ABT_JOINT_COUNT; ++i)
				{
					c2[i] = 0.0;
				}

			}

			// capture calibrated body motion
			std::vector<SimTK::Vec3> v1(K4ABT_JOINT_COUNT), v2(K4ABT_JOINT_COUNT);
			for (int i = 0; i < (int)K4ABT_JOINT_COUNT; ++i)
			{
				v1[i] = main_body.skeleton.joints[i].position.v;
				v2[i] = secondary_body.skeleton.joints[i].position.v;

				c1[i] = confidence_level(main_body.skeleton.joints[i]);
				c2[i] = confidence_level(secondary_body.skeleton.joints[i]);
			}

			// insert into right coordinates
			for (int i = 0; i < v1.size(); ++i)
			{
				const double x1 = v1[i][0];
				const double y1 = v1[i][1];
				const double z1 = v1[i][2];

				v1[i][0] = z1 / 1e3;
				v1[i][1] = -y1 / 1e3 + 1.;
				v1[i][2] = x1 / 1e3;			// mm to m
			}
			for (int i = 0; i < v2.size(); ++i)
			{
				const double x2 = v2[i][0];
				const double y2 = v2[i][1];
				const double z2 = v2[i][2];

				v2[i][0] = z2 / 1e3;
				v2[i][1] = -y2 / 1e3 + 1.;
				v2[i][2] = x2 / 1e3;			// mm to m
			}

			tstv1.appendRow(frame1.get_device_timestamp().count() / 1e6, v1);
			tstv2.appendRow(frame2.get_device_timestamp().count() / 1e6, v2);

			tst1.appendRow(frame1.get_device_timestamp().count() / 1e6, c1);
			tst2.appendRow(frame2.get_device_timestamp().count() / 1e6, c2);		// us to s
		}
		else
		{
			//  It should never hit timeout when K4A_WAIT_INFINITE is set.
			std::cout << "Error! No body recognized in this frame! Time out!" << std::endl;
			continue;
		}

		frame1.reset();
		frame2.reset();
	}
	std::cout << "Capture end." << std::endl;
	tracker1.shutdown();
	tracker2.shutdown();
	tracker1.destroy();
	tracker2.destroy();
	playback1.close();
	playback2.close();

	/*
	https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/TRCFileAdapter.h
	https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/TRCFileAdapter.cpp
	*/
	
	const double dr = 1. / 30.; // 30 fps
	const int nf = tstv1.getNumRows();

	tstv1.addTableMetaData("header", SimTK::String("PathFileType\t4\t(X/Y/Z)\tsample.trc"));
	tstv1.addTableMetaData("DataRate", SimTK::String(dr));
	tstv1.addTableMetaData("CameraRate", SimTK::String(dr));
	tstv1.addTableMetaData("NumFrames", SimTK::String(nf));
	tstv1.addTableMetaData("NumMarkers", SimTK::String(K4ABT_JOINT_COUNT));
	tstv1.addTableMetaData("Units", SimTK::String("m"));
	tstv1.addTableMetaData("OrigDataRate", SimTK::String(dr));
	tstv1.addTableMetaData("OrigDataStartFrame", SimTK::String(1));
	tstv1.addTableMetaData("OrigNumFrames", SimTK::String(nf));

	const double zt1 = tstv1.getIndependentColumn()[0];
	for (int i = 0; i < tstv1.getIndependentColumn().size(); ++i)
		const_cast<std::vector<double>*> (&tstv1.getIndependentColumn())->at(i) -= zt1;

	tstv2.addTableMetaData("header", SimTK::String("PathFileType\t4\t(X/Y/Z)\tsample.trc"));
	tstv2.addTableMetaData("DataRate", SimTK::String(dr));
	tstv2.addTableMetaData("CameraRate", SimTK::String(dr));
	tstv2.addTableMetaData("NumFrames", SimTK::String(nf));
	tstv2.addTableMetaData("NumMarkers", SimTK::String(K4ABT_JOINT_COUNT));
	tstv2.addTableMetaData("Units", SimTK::String("m"));
	tstv2.addTableMetaData("OrigDataRate", SimTK::String(dr));
	tstv2.addTableMetaData("OrigDataStartFrame", SimTK::String(1));
	tstv2.addTableMetaData("OrigNumFrames", SimTK::String(nf));

	const double zt2 = tstv2.getIndependentColumn()[0];
	for (int i = 0; i < tstv2.getIndependentColumn().size(); ++i)
		const_cast<std::vector<double>*> (&tstv2.getIndependentColumn())->at(i) -= zt2;
	

	OpenSim::TRCFileAdapter trcfa1;
	trcfa1.write(tstv1, "motion1_raw.trc");
	OpenSim::TRCFileAdapter trcfa2;
	trcfa2.write(tstv2, "motion2_raw.trc");
}

/**
\brief Converts and merges two synched mkv files to one trc file (based on confidence level of markers).
\param[in] fnMKV the name of the Matroska file.
\param[in] fnTRC the name of the trc file.
*/
void capture(const std::string& fnMKV1, const std::string& fnMKV2, const std::string& fnTRC) {

	OpenSim::TimeSeriesTableVec3 tstv1, tstv2, tstv;
	tstv1.setColumnLabels({
		"PELVIS",
		"SPINE_NAVEL",
		"SPINE_CHEST",
		"NECK",
		"CLAVICLE_LEFT",
		"SHOULDER_LEFT",
		"ELBOW_LEFT",
		"WRIST_LEFT",
		"HAND_LEFT",
		"HANDTIP_LEFT",
		"THUMB_LEFT",
		"CLAVICLE_RIGHT",
		"SHOULDER_RIGHT",
		"ELBOW_RIGHT",
		"WRIST_RIGHT",
		"HAND_RIGHT",
		"HANDTIP_RIGHT",
		"THUMB_RIGHT",
		"HIP_LEFT",
		"KNEE_LEFT",
		"ANKLE_LEFT",
		"FOOT_LEFT",
		"HIP_RIGHT",
		"KNEE_RIGHT",
		"ANKLE_RIGHT",
		"FOOT_RIGHT]",
		"HEAD",
		"NOSE",
		"EYE_LEFT",
		"EAR_LEFT",
		"EYE_RIGHT",
		"EAR_RIGHT"
	});
	tstv2.setColumnLabels(tstv1.getColumnLabels());
	tstv.setColumnLabels(tstv1.getColumnLabels());

	OpenSim::TimeSeriesTable tst1, tst2;
	tst1.setColumnLabels(tstv1.getColumnLabels());
	tst2.setColumnLabels(tstv1.getColumnLabels());

	tracking(fnMKV1, fnMKV2, tstv1, tstv2, tst1, tst2);
	
	// merge both motions to one based on confidence level
	int NumRows = tst1.getNumRows();
	int NumColumns = tst1.getNumColumns();
	std::cout << "Merging start." << std::endl;

	for (int i = 0; i < NumRows; ++i) {

		SimTK::RowVector_<SimTK::Vec3> row(NumColumns);

		for (int j = 0; j < NumColumns; ++j) {

			std::string label = tst1.getColumnLabel(j);

			double c1 = tst1.getDependentColumn(label)[i];
			double c2 = tst2.getDependentColumn(label)[i];

			SimTK::Vec3 v1 = tstv1.getDependentColumn(label)[i];
			SimTK::Vec3 v2 = tstv2.getDependentColumn(label)[i];

			if (c1 == c2) { // equal confidence
				for (int k = 0; k < 3; ++k) {
					row[j][k] = (v1[k] + v2[k]) / 2;
				}
			}
			else { // better confidence

				if (c1 == 0.0) row[j] = v2;
				else if (c2 == 0.0) row[j] = v1;
				else if (c1 > c2) {
					for (int k = 0; k < 3; ++k) {
						row[j][k] = (2.0 * v1[k] + 1.0 * v2[k]) / 3.0;
					}
				}
				else {
					for (int k = 0; k < 3; ++k) {
						row[j][k] = (2.0 * v2[k] + 1.0 * v1[k]) / 3.0;
					}
				}
			} 
		}
		tstv.appendRow(tstv1.getIndependentColumn()[i], row);
	}

	std::cout << "Merging end." << std::endl;

	const double dr = 1. / 30.; // 30 fps
	const int nf = tstv.getNumRows();

	tstv.addTableMetaData("header", SimTK::String("PathFileType\t4\t(X/Y/Z)\tsample.trc"));
	tstv.addTableMetaData("DataRate", SimTK::String(dr));
	tstv.addTableMetaData("CameraRate", SimTK::String(dr));
	tstv.addTableMetaData("NumFrames", SimTK::String(nf));
	tstv.addTableMetaData("NumMarkers", SimTK::String(K4ABT_JOINT_COUNT));
	tstv.addTableMetaData("Units", SimTK::String("m"));
	tstv.addTableMetaData("OrigDataRate", SimTK::String(dr));
	tstv.addTableMetaData("OrigDataStartFrame", SimTK::String(1));
	tstv.addTableMetaData("OrigNumFrames", SimTK::String(nf));

	OpenSim::TRCFileAdapter trcfa;
	trcfa.write(tstv, fnTRC);
}

/** 
\brief Smoothens motion tracked in trc file.
\param[in] fnTRCin the name of the noisy trc file
\param[in] fnTRCout the name of the smooth trc file
*/
void filter(const std::string& fnTRCin, const std::string& fnTRCout, const double eps) {
	std::cout << "Smoothing start." << std::endl;

	OpenSim::TRCFileAdapter trcfaIN;
	auto adt = trcfaIN.getDataTable(trcfaIN.read(fnTRCin), trcfaIN._markers);
	adt->setColumnLabels({
		"PELVIS",
		"SPINE_NAVEL",
		"SPINE_CHEST",
		"NECK",
		"CLAVICLE_LEFT",
		"SHOULDER_LEFT",
		"ELBOW_LEFT",
		"WRIST_LEFT",
		"HAND_LEFT",
		"HANDTIP_LEFT",
		"THUMB_LEFT",
		"CLAVICLE_RIGHT",
		"SHOULDER_RIGHT",
		"ELBOW_RIGHT",
		"WRIST_RIGHT",
		"HAND_RIGHT",
		"HANDTIP_RIGHT",
		"THUMB_RIGHT",
		"HIP_LEFT",
		"KNEE_LEFT",
		"ANKLE_LEFT",
		"FOOT_LEFT",
		"HIP_RIGHT",
		"KNEE_RIGHT",
		"ANKLE_RIGHT",
		"FOOT_RIGHT]",
		"HEAD",
		"NOSE",
		"EYE_LEFT",
		"EAR_LEFT",
		"EYE_RIGHT",
		"EAR_RIGHT"
		});
	auto& tstv = *(std::dynamic_pointer_cast<OpenSim::TimeSeriesTableVec3>(adt));

	OpenSim::TimeSeriesTable tst = tstv.flatten();

	OpenSim::GCVSplineSet GSVspline(tst, tst.getColumnLabels(), 3, eps);

	SimTK::Array_<SimTK::Vector> splines(tst.getNumColumns());
	for (int i = 0; i < tst.getNumColumns(); ++i) {
		SimTK::Vector v(tstv.getNumRows());
		for (int j = 0; j < tstv.getNumRows(); ++j)
			v[j] = GSVspline.evaluate(i, 0, tstv.getIndependentColumn()[j]);
		splines[i] = v;
	}

	OpenSim::TimeSeriesTableVec3 tstvOUT;
	tstvOUT.setColumnLabels(tstv.getColumnLabels());

	for (int i = 0; i < tstv.getNumRows(); ++i) {
		SimTK::RowVector_<SimTK::Vec3> row(tstv.getNumColumns());
		for (int j = 0; j < tstv.getNumColumns(); ++j)
			row[j] = SimTK::Vec3(splines[3*j+0][i], splines[3*j+1][i], splines[3*j+2][i]);
		tstvOUT.appendRow(tstv.getIndependentColumn()[i], row);
	}

	std::cout << "Smoothing end." << std::endl;

	const double dr = 1. / 30.; // 30 fps
	const int nf = tstvOUT.getNumRows();

	tstvOUT.addTableMetaData("header", SimTK::String("PathFileType\t4\t(X/Y/Z)\tsample.trc"));
	tstvOUT.addTableMetaData("DataRate", SimTK::String(dr));
	tstvOUT.addTableMetaData("CameraRate", SimTK::String(dr));
	tstvOUT.addTableMetaData("NumFrames", SimTK::String(nf));
	tstvOUT.addTableMetaData("NumMarkers", SimTK::String(K4ABT_JOINT_COUNT));
	tstvOUT.addTableMetaData("Units", SimTK::String("m"));
	tstvOUT.addTableMetaData("OrigDataRate", SimTK::String(dr));
	tstvOUT.addTableMetaData("OrigDataStartFrame", SimTK::String(1));
	tstvOUT.addTableMetaData("OrigNumFrames", SimTK::String(nf));

	OpenSim::TRCFileAdapter trcfaOUT;
	trcfaOUT.write(tstvOUT, fnTRCout);
}

/**
\brief Converts trc file to sto file.
\param[in] model the Model.
\param[in] fnTRC the name of the trc file.
\param[in] fnSTO the name of the sto file.
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Tools/InverseKinematicsTool.h
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Tools/InverseKinematicsTool.cpp
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Simulation/InverseKinematicsSolver.h
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Simulation/InverseKinematicsSolver.cpp
*/
void inverseKinematics(OpenSim::Model model, const std::string& fnTRC, const std::string& fnSTO) {

	model.addMarker(new OpenSim::Marker("PELVIS", model.getBodySet().get("Pelvis"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("SPINE_CHEST", model.getBodySet().get("Thorax"), 
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("HIP_RIGHT", model.getBodySet().get("Thigh_r"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("KNEE_RIGHT", model.getBodySet().get("LowerLeg_r"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("ANKLE_RIGHT", model.getBodySet().get("LowerLeg_r"),
		SimTK::Vec3(0., .42, 0.)));
	model.addMarker(new OpenSim::Marker("HIP_LEFT", model.getBodySet().get("Thigh_l"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("KNEE_LEFT", model.getBodySet().get("LowerLeg_l"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("ANKLE_LEFT", model.getBodySet().get("LowerLeg_l"),
		SimTK::Vec3(0., .42, 0.)));
	model.addMarker(new OpenSim::Marker("SHOULDER_RIGHT", model.getBodySet().get("UpperArm_r"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("ELBOW_RIGHT", model.getBodySet().get("Forearm_r"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("WRIST_RIGHT", model.getBodySet().get("Forearm_r"),
		SimTK::Vec3(0., .25, 0.)));
	model.addMarker(new OpenSim::Marker("SHOULDER_LEFT", model.getBodySet().get("UpperArm_l"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("ELBOW_LEFT", model.getBodySet().get("Forearm_l"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("WRIST_LEFT", model.getBodySet().get("Forearm_l"),
		SimTK::Vec3(0., .25, 0.)));
	model.addMarker(new OpenSim::Marker("SPINE_NAVEL", model.getBodySet().get("Lumbar"),
		SimTK::Vec3(0., 0., 0.)));
	model.addMarker(new OpenSim::Marker("HEAD", model.getBodySet().get("Head"),
		SimTK::Vec3(0., 0., 0.)));

	OpenSim::TRCFileAdapter trcfa;
	auto tstv = std::dynamic_pointer_cast<OpenSim::TimeSeriesTableVec3>(
		trcfa.getDataTable(trcfa.read(fnTRC), trcfa._markers));

	auto state = model.initSystem();
	auto assembler = std::unique_ptr<SimTK::Assembler>(new SimTK::Assembler(model.getMultibodySystem()));
	auto markers = new SimTK::Markers();
	assembler->adoptAssemblyGoal(markers);

	for (int i = 0; i < model.getNumMarkers(); ++i) {
		const auto& m = model.getMarkerSet()[i];
		markers->addMarker(m.getName(), m.getParentFrame().getMobilizedBody(),
			m.getParentFrame().findTransformInBaseFrame() * m.get_location(), 1.);
	}
	SimTK::Array_<std::string> oo;
	for (int i = 0; i < markers->getNumMarkers(); ++i)
		oo.emplace_back(markers->getMarkerName((SimTK::Markers::MarkerIx)i));
	markers->defineObservationOrder(oo);

	auto storage = std::unique_ptr<OpenSim::Storage>(new OpenSim::Storage(tstv->getNumRows()));
	storage->setOutputFileName(fnSTO);
	OpenSim::Array<std::string> labels;
	labels.append("time");
	for (int i = 0; i < model.getNumCoordinates(); ++i)
		labels.append(model.getCoordinateSet().get(i).getAbsolutePathString() + "/value");
	storage->setColumnLabels(labels);

	SimTK::Array_<SimTK::Vec3> observations(model.getNumMarkers());
	for (int i = 0; i < tstv->getNumRows(); ++i) {
		observations[0] = tstv->getRowAtIndex(i)[K4ABT_JOINT_PELVIS];
		observations[1] = tstv->getRowAtIndex(i)[K4ABT_JOINT_SPINE_CHEST];
		observations[2] = tstv->getRowAtIndex(i)[K4ABT_JOINT_HIP_RIGHT];
		observations[3] = tstv->getRowAtIndex(i)[K4ABT_JOINT_KNEE_RIGHT];
		observations[4] = tstv->getRowAtIndex(i)[K4ABT_JOINT_ANKLE_RIGHT];
		observations[5] = tstv->getRowAtIndex(i)[K4ABT_JOINT_HIP_LEFT];
		observations[6] = tstv->getRowAtIndex(i)[K4ABT_JOINT_KNEE_LEFT];
		observations[7] = tstv->getRowAtIndex(i)[K4ABT_JOINT_ANKLE_LEFT];
		observations[8] = tstv->getRowAtIndex(i)[K4ABT_JOINT_SHOULDER_RIGHT];
		observations[9] = tstv->getRowAtIndex(i)[K4ABT_JOINT_ELBOW_RIGHT];
		observations[10] = tstv->getRowAtIndex(i)[K4ABT_JOINT_WRIST_RIGHT];
		observations[11] = tstv->getRowAtIndex(i)[K4ABT_JOINT_SHOULDER_LEFT];
		observations[12] = tstv->getRowAtIndex(i)[K4ABT_JOINT_ELBOW_LEFT];
		observations[13] = tstv->getRowAtIndex(i)[K4ABT_JOINT_WRIST_LEFT];
		observations[14] = tstv->getRowAtIndex(i)[K4ABT_JOINT_SPINE_NAVEL];
		observations[15] = tstv->getRowAtIndex(i)[K4ABT_JOINT_HEAD];

		markers->moveAllObservations(observations);
		if (i == 0) assembler->assemble(state);
		assembler->track();
		assembler->updateFromInternalState(state);
		model.realizePosition(state);

		SimTK::Vector v(model.getNumCoordinates(), 0.);
		for (int j = 0; j < model.getNumCoordinates(); ++j) {
			const auto& c = model.getCoordinateSet().get(j);
			v[j] = c.getValue(state);
		}
		OpenSim::StateVector sv(tstv->getIndependentColumn().at(i), v);
		storage->append(sv);
	}

	storage->print(fnSTO);
}

/**
\brief Shows the movement.
\param[in] model the Model.
\param[in] fnSTO the name of the sto file.
*/
void forwardKinematics(OpenSim::Model model, const std::string& fnSTO) {
	auto storage = std::unique_ptr<OpenSim::Storage>(new OpenSim::Storage(fnSTO));
	model.setUseVisualizer(true);
	auto state = model.initSystem();
	for (int i = 0; i < storage->getSize(); ++i) {
		const auto& sv = storage->getStateVector(i)->getData();
		for (int j = 0; j < model.getNumCoordinates(); ++j) {
			auto& c = model.updCoordinateSet().get(j);
			double d = sv[storage->getStateIndex(c.getAbsolutePathString() + "/value")];
			if (storage->isInDegrees() && (c.getMotionType() == OpenSim::Coordinate::Rotational))
				d = SimTK::convertDegreesToRadians(d);
			c.setValue(state, d);
		}
		model.realizePosition(state);
		model.getVisualizer().show(state);
		std::this_thread::sleep_for(std::chrono::milliseconds((long long)(1. / 30. * 1e3)));
	}
}

/**
\brief Forward Kinematics to simulate the movement in a zero gravity setting.
\param[in] model the Model.
\param[in] inputSTOfn the name of the input file.
\param[in] outputSTOfn the name of the output file.
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/Storage.h
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/Storage.cpp
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/Signal.h
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Common/Signal.cpp
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Tools/CorrectionController.h
\sa https://github.com/opensim-org/opensim-core/blob/main/OpenSim/Tools/CorrectionController.cpp
*/
void simulation(OpenSim::Model model, const std::string& inputSTOfn, const std::string& outputSTOfn) {
	auto inputSTO = std::unique_ptr<OpenSim::Storage>(new OpenSim::Storage(inputSTOfn));
	if (inputSTO->isInDegrees())
		model.getSimbodyEngine().convertDegreesToRadians(*inputSTO);
	const double firstTime = inputSTO->getFirstTime();
	const double lastTime = inputSTO->getLastTime();

	inputSTO->pad(inputSTO->getSize() / 2);
	for (int i = 0; i < inputSTO->getColumnLabels().getSize() - 1; ++i) { // i is StateIndex
		OpenSim::Array<double> d(0., inputSTO->getSize());
		inputSTO->getDataColumn(i, d);
		for (int j = 0; j < inputSTO->findIndex(firstTime); ++j)
			d[j] = d[inputSTO->findIndex(firstTime)];
		for (int j = inputSTO->findIndex(lastTime) + 1; j < inputSTO->getSize(); ++j)
			d[j] = d[inputSTO->findIndex(lastTime)];
		inputSTO->setDataColumn(i, d);
	}
	inputSTO->lowpassIIR(1.);
	inputSTO->crop(firstTime, lastTime);

	OpenSim::GCVSplineSet inputGCVSS(5, inputSTO.get());

	auto statezero = model.initSystem(); // first initialize system

	auto outputSTO = std::unique_ptr<OpenSim::Storage>(new OpenSim::Storage());
	outputSTO->setOutputFileName(outputSTOfn);
	OpenSim::Array<std::string> labels;
	labels.append("time");
	for (int i = 0; i < model.getNumCoordinates(); ++i)
		labels.append(model.getCoordinateSet().get(i).getAbsolutePathString() + "/value");
	outputSTO->setColumnLabels(labels);

	model.setGravity(SimTK::Vec3(0., 0., 0.));

	auto controller = new OpenSim::PrescribedController();
	model.addController(controller);
	for (int i = 0; i < model.getNumCoordinates(); ++i) {
		const auto& c = model.getCoordinateSet().get(i);

		OpenSim::CoordinateActuator* ca;
		if (model.getActuators().contains(c.getName() + "_ca")) {
			std::cout << "Using the existing " << c.getName() << "_ca." << std::endl;
			ca = OpenSim::CoordinateActuator::safeDownCast(
				&(model.updActuators().get(c.getName() + "_ca")));
			controller->addActuator(*ca);
		}
		else {
			std::cout << "Adding a new " << c.getName() << "_ca." << std::endl;
			ca = new OpenSim::CoordinateActuator(c.getName());
			ca->setName(c.getName() + "_ca");
			ca->setOptimalForce(1.);
			ca->setMinControl(-1e3);
			ca->setMaxControl(1e3);
			model.addForce(ca);
			controller->addActuator(*ca);
		}
		controller->prescribeControlForActuator(ca->getName(), new OpenSim::Constant(0.));
	}

	model.setUseVisualizer(true);
	auto state = model.initSystem();

	state.setTime(firstTime);
	for (int i = 0; i < model.getNumCoordinates(); ++i) {
		auto& c = model.updCoordinateSet().get(i);

		c.setValue(state, inputGCVSS.evaluate(
			inputGCVSS.getIndex(c.getAbsolutePathString() + "/value"), 0, firstTime));
		c.setSpeedValue(state, 0.);

		if (false
			|| (c.getName() == "SpineL_LL")
			|| (c.getName() == "SpineL_RR")
			|| (c.getName() == "SpineL_FE")
			|| (c.getName() == "SpineU_FE")
			|| (c.getName() == "Hip_EF_r")
			|| (c.getName() == "Hip_RR_r")
			|| (c.getName() == "Hip_DB_r")
			|| (c.getName() == "Knee_FE_r")
			|| (c.getName() == "Hip_EF_l")
			|| (c.getName() == "Hip_RR_l")
			|| (c.getName() == "Hip_BD_l")
			|| (c.getName() == "Knee_FE_l")
			|| (c.getName() == "Shoulder_BD_r")
			|| (c.getName() == "Shoulder_EF_r")
			|| (c.getName() == "Shoulder_RR_r")
			|| (c.getName() == "Elbow_EF_r")
			|| (c.getName() == "Shoulder_DB_l")
			|| (c.getName() == "Shoulder_EF_l")
			|| (c.getName() == "Shoulder_RR_l")
			|| (c.getName() == "Elbow_EF_l")
			|| (c.getName() == "Neck_LL")
			|| (c.getName() == "Neck_RR")
			|| (c.getName() == "Neck_FE")
			) {
		}
		else {
			for (const auto& x : model.getComponentList<OpenSim::CoordinateLimitForce>()) {
				if (x.get_coordinate() == c.getName()) {
					if (c.getValue(state) > SimTK::convertDegreesToRadians(x.getUpperLimit()))
						c.setValue(state, SimTK::convertDegreesToRadians(x.getUpperLimit()));
					if (c.getValue(state) < SimTK::convertDegreesToRadians(x.getLowerLimit()))
						c.setValue(state, SimTK::convertDegreesToRadians(x.getLowerLimit()));
					break;
				}
			}
		}
	} // To avoid spring-like movement.

	std::cout << "Simulation start." << std::endl;
	for (int i = inputSTO->findIndex(firstTime) + 1; i <= inputSTO->findIndex(lastTime); ++i) {
		double t_;
		double _t;
		inputSTO->getTime(i - 1, t_);
		inputSTO->getTime(i, _t);
		const int n = 100;
		for (int j = 0; j < n; ++j) {
			const double t = t_ + (_t - t_) / ((double)n) * ((double)j);
			std::vector<double> cs(model.getNumCoordinates());
			for (int k = 0; k < model.getNumCoordinates(); ++k) {
				const auto& c = model.getCoordinateSet().get(k);
				if (false
					|| (c.getName() == "SpineL_LL")
					|| (c.getName() == "SpineL_RR")
					|| (c.getName() == "SpineL_FE")
					|| (c.getName() == "SpineU_FE")
					|| (c.getName() == "Hip_EF_r")
					|| (c.getName() == "Hip_RR_r")
					|| (c.getName() == "Hip_DB_r")
					|| (c.getName() == "Knee_FE_r")
					|| (c.getName() == "Hip_EF_l")
					|| (c.getName() == "Hip_RR_l")
					|| (c.getName() == "Hip_BD_l")
					|| (c.getName() == "Knee_FE_l")
					|| (c.getName() == "Shoulder_BD_r")
					|| (c.getName() == "Shoulder_EF_r")
					|| (c.getName() == "Shoulder_RR_r")
					|| (c.getName() == "Elbow_EF_r")
					|| (c.getName() == "Shoulder_DB_l")
					|| (c.getName() == "Shoulder_EF_l")
					|| (c.getName() == "Shoulder_BD_r")
					|| (c.getName() == "Shoulder_RR_r")
					|| (c.getName() == "Elbow_EF_l")
					|| (c.getName() == "Neck_LL")
					|| (c.getName() == "Neck_RR")
					|| (c.getName() == "Neck_FE")
					) {
					if (!inputGCVSS.contains(c.getAbsolutePathString() + "/value")) continue;
					auto ca = OpenSim::CoordinateActuator::safeDownCast(
						&(model.updActuators().get(c.getName() + "_ca")));
					const auto si = inputGCVSS.getIndex(c.getAbsolutePathString() + "/value");
					const double kp = 100.;
					const double kv = 20.;
					controller->prescribeControlForActuator(
						controller->getActuatorSet().get(c.getName() + "_ca").getName(),
						new OpenSim::Constant(
							- (kp / ca->getOptimalForce())
								* (c.getValue(state) - inputGCVSS.evaluate(si, 0, t))
							- (kv / ca->getOptimalForce())
								* (0. - inputGCVSS.evaluate(si, 1, t))));
				}
			}

			auto manager = std::unique_ptr<OpenSim::Manager>(new OpenSim::Manager(model, state)); // Probably, Manager has to be in the heap, since it holds a Storage.
			state = manager->integrate(t);
		}

		SimTK::Vector v(model.getNumCoordinates(), 0.);
		for (int j = 0; j < model.getNumCoordinates(); ++j)
			v[j] = model.getCoordinateSet().get(j).getValue(state);
		outputSTO->append(OpenSim::StateVector(state.getTime(), v));
	}
	std::cout << "Simulation end." << std::endl;
	outputSTO->print(outputSTOfn);

}

/**
\brief Starts the program.
\param[in] argc the size of the array of the command line arguments.
\param[in] argv the array of the command line arguments.
\return the exit status.
*/
int main(int argc, char* argv[]) {
	capture("record1.mkv", "record2.mkv", "motion_raw.trc");
	filter("motion_raw.trc", "motion.trc", 0.0001);

	auto model = OpenSim::Model("model.osim"); // without Muscle
	inverseKinematics(model, "motion.trc", "motion.sto");
	forwardKinematics(model, "motion.sto"); 

	simulation(model, "motion.sto", "output.sto");

	return EXIT_SUCCESS;
}
