/**
\file transfrom.h
\brief Transforms between the coordinate systems of two tracked bodies based on MOCAP
\author Adelina Baerligea
\date 2023/10/27
\sa MOCAP, https://github.com/andyj1/kinect
\sa Microsoft Azure Kinect DK, https://learn.microsoft.com/en-us/azure/kinect-dk/
\sa Azure Kinect Body Tracking SDK
\sa OpenCV 4.7.0, https://opencv.org/
\sa CMake, https://cmake.org
\sa Doxygen, https://www.doxygen.nl/
*/
#pragma once

#include <opencv2/core.hpp>

void transform_body(k4abt_body_t& main_body, k4abt_body_t& secondary_body);
void arun(cv::Mat& streamfrom, cv::Mat& streamto, cv::Mat& R, cv::Mat& T);

/**
\brief Transforms secondary to main body space coordinates.
\param[in] main_body the tracked body captured by the main device.
\param[in, out] secondary_body the tracked body captured by the secondary device.
\sa https://github.com/andyj1/kinect/blob/master/src/two.cpp
\sa https://github.com/andyj1/kinect/blob/master/src/two.hpp
*/
void transform_body(k4abt_body_t& main_body, k4abt_body_t& secondary_body)
{
	// using Arun's method for computing Rotation and Translation Matrices from positions (for 3dof) / orientations (for 6dof)
	// paper: https://doi.org/10.1109/TPAMI.1987.4767965 (Least-Squares Fitting of Two 3-D Point Sets)

	// for each joint object
	cv::Mat main(3, K4ABT_JOINT_COUNT, CV_32F), secondary(3, K4ABT_JOINT_COUNT, CV_32F);
	for (int joint = 0; joint < (int)K4ABT_JOINT_COUNT; joint++)
	{
		main.at<float>(0, joint) = main_body.skeleton.joints[joint].position.v[0];
		main.at<float>(1, joint) = main_body.skeleton.joints[joint].position.v[1];
		main.at<float>(2, joint) = main_body.skeleton.joints[joint].position.v[2];
		secondary.at<float>(0, joint) = secondary_body.skeleton.joints[joint].position.v[0];
		secondary.at<float>(1, joint) = secondary_body.skeleton.joints[joint].position.v[1];
		secondary.at<float>(2, joint) = secondary_body.skeleton.joints[joint].position.v[2];
	}

	// ====Apply transformation====
	// R, T from secondary to main coordinate space
	cv::Mat R, T;
	arun(secondary, main, R, T); // R: [3x3], T: [1x3]

	// main, secondary: [3 x 32]
	cv::Mat T_rep;
	cv::repeat(T, 1, K4ABT_JOINT_COUNT, T_rep); // T_rep: [3x32]

	cv::Mat secondary_tf;
	cv::add(R * secondary, T_rep, secondary_tf);

	// reassign body object with transformed positions
	for (int joint = 0; joint < (int)K4ABT_JOINT_COUNT; joint++)
	{
		secondary_body.skeleton.joints[joint].position.v[0] = secondary_tf.at<float>(0, joint);
		secondary_body.skeleton.joints[joint].position.v[1] = secondary_tf.at<float>(1, joint);
		secondary_body.skeleton.joints[joint].position.v[2] = secondary_tf.at<float>(2, joint);
	}
}

/**
\brief Implementation of Arun's method for Least Squares Point Set Fitting
\sa https://doi.org/10.1109/TPAMI.1987.4767965
*/
void arun(cv::Mat& streamfrom, cv::Mat& streamto, cv::Mat& R, cv::Mat& T)
{
	// find mean across the row (all joints for each x,y,z)
	cv::Mat avg_streamfrom, avg_streamto;							// main, secondary: [3x32]
	reduce(streamfrom, avg_streamfrom, 1, cv::REDUCE_AVG);          // avg_main: [3x1]
	reduce(streamto, avg_streamto, 1, cv::REDUCE_AVG);

	// find deviations from the mean
	cv::Mat rep_avg_streamfrom, rep_avg_streamto;
	cv::repeat(avg_streamfrom, 1, K4ABT_JOINT_COUNT, rep_avg_streamfrom);      // rep_avg_main: [3x32]
	cv::repeat(avg_streamto, 1, K4ABT_JOINT_COUNT, rep_avg_streamto);

	cv::Mat streamfrom_sub, streamto_sub;
	subtract(streamfrom, rep_avg_streamfrom, streamfrom_sub);
	subtract(streamto, rep_avg_streamto, streamto_sub);

	// take singular value decomposition and compute R and T matrices
	cv::Mat s, u, v;
	cv::SVDecomp(streamfrom_sub * streamto_sub.t(), s, u, v);
	v = v.t();
	R = v * u.t();
	double det = cv::determinant(R);

	if (det >= 0)
	{
		subtract(avg_streamto, (R * avg_streamfrom), T);
	}
	else {
		//T = cv::Mat(3, 1, CV_32F, cv::Scalar(1, 1, 1));
		throw std::domain_error("Warning: Determinant below 0! Tranform failed!");
	}
	// R: [3x3], T: [3x1];
}